import os, sys, re
# Relative Imports
TEST_DIR = os.path.dirname(__file__)
PARENT_DIR = os.path.dirname(TEST_DIR)
sys.path.insert(0, PARENT_DIR)
from brukeropus import read_opus, find_opus_files, OPUSFile
from brukeropus.file.utils import _print_centered
from brukeropus.file.constants import PARAM_LABELS


def get_all_blocks(opusfile: OPUSFile) -> list:
    '''Returns a list of all `FileBlock` in an `OPUSFile` instance.'''
    blocks = [opusfile.directory.block] + opusfile.parse_error_blocks + opusfile.unknown_blocks + \
        opusfile.unmatched_data_blocks + opusfile.unmatched_data_status_blocks
    if hasattr(opusfile, 'params'):
        blocks = blocks + opusfile.params.blocks
    if hasattr(opusfile, 'rf_params'):
        blocks = blocks + opusfile.rf_params.blocks
    for d in opusfile.iter_all_data():
        blocks = blocks + d.blocks
    for report in opusfile.reports:
        blocks.append(report.block)
    if opusfile.history:
        blocks = blocks + opusfile.history_blocks
    blocks = [b for b in blocks if b is not None]
    return blocks


def filter_data_with_mismatched_num_blocks(data: list) -> list:
    '''Filters a list of `OPUSFile` to include only instances where `num_blocks` from the header does not match the
    length of the parsed blocks in the directory and the length of all sorted blocks in the `OPUSFile`'''
    data = [d for d in data if d]
    mismatched = []
    for o in data:
        blocks = get_all_blocks(o)
        try:
            if not (len(blocks) == len(o.directory.block.data) and len(blocks) == o.directory.num_blocks):
                mismatched.append(o)
        except:
            print('Directory block missing for file (skipping mismatched block test):', o.filepath)
    return mismatched


def filter_data_with_unknown_blocks(data: list) -> list:
    '''Filters a list of `OPUSFile` to include only instances where there are unknown/uncategorized blocks.  This
    function ignores blocks of type: (0, 0, 0, 0, 0, 0) because OPUS also appears to ignore those blocks.'''
    data = [d for d in data if d]
    unknown = [d for d in data if len(d.unknown_blocks) > 0]
    return [d for d in unknown if any(b.type != (0, 0, 0, 0, 0, 0) for b in d.unknown_blocks)]


def filter_data_that_failed_to_parse(data: list) -> list:
    '''Filters a list of `OPUSFile` to include only instances where there are blocks whose `bytes` attribute were not
    set to zero or whose `parser` attribute has not been set (indicates parsing was not attempted or there was an
    exception during parsing).'''
    data = [d for d in data if d]
    bad_parse = []
    for o in data:
        blocks = get_all_blocks(o)
        blocks = [b for b in blocks if b.type != (0, 0, 0, 0, 0, 0)]
        if any(b.parser is None or b.bytes != b'' for b in blocks):
            bad_parse.append(o)
    return bad_parse


def find_all_files(directory: str) -> list:
    '''Recursively finds all files (regardless of filetype) in a directory.'''
    filepaths = []
    for root, dirs, filenames in os.walk(directory):
        filepaths = filepaths + [os.path.join(root, f) for f in filenames]
    return filepaths


def print_pass(message):
    'Prints a formatted pass test message'
    print('| Passed |  ' + message)


def print_fail(message):
    'Prints a formatted failed test message'
    print('| FAILED |  ' + message)


if __name__ == "__main__":
    # ==============================================================================================
    # Test Report Settings
    directory = os.path.join(TEST_DIR, 'files')  # Directory with test files
    width = 120  # Width of report in characters
    print_non_opus_ext = False  # Whether to print all files without OPUS numeric extension
    # ----------------------------------------------------------------------------------------------
    # Initialize Data
    opus_files = find_opus_files(directory, recursive=True)
    opus_data = [read_opus(f) for f in opus_files]
    for o in opus_data:
        o.rel_path = '\\' + os.path.relpath(o.filepath, directory)  # Add rel_path attribute for printing
    # ----------------------------------------------------------------------------------------------
    # Print header
    print()
    print('=' * width)
    _print_centered('Test File Directory: ' + str(directory), width=width)
    # ----------------------------------------------------------------------------------------------
    # Find all files and check for presence of non-opus files (based on extension and reading bytes)
    all_files = find_all_files(directory)
    non_opus_ext = [f for f in all_files if f not in opus_files]
    opus_ext_but_not_opus = [d for d in opus_data if not d]
    print(len(all_files), 'files found')
    print(len(opus_files), 'files with OPUS numeric extension.')
    if len(non_opus_ext) > 0 and print_non_opus_ext:
        print('Files without numeric extension:')
        for f in non_opus_ext:
            print('   ', '\\' + os.path.relpath(f, directory))
    if len(opus_ext_but_not_opus) > 0:
        print('Files with numeric extension that are not valid OPUS files:')
        for o in opus_ext_but_not_opus:
            print('   ', o.rel_path)
    # ----------------------------------------------------------------------------------------------
    # Check that parsed directory block and sorted blocks in OPUSFile match in length and agree with
    #   num_blocks value from header
    mismatched = filter_data_with_mismatched_num_blocks(opus_data)
    print('.' * width)
    if len(mismatched) == 0:
        print_pass('All file blocks were found and consistent (' + str(len(opus_data)) + ' files)')
    else:
        print_fail('Some mismatched block numbers: ' + str(len(mismatched)) + '/' + str(len(opus_data)) + ' files')
        for o in mismatched:
            print('\n', o.rel_path,
                  '\n    num_blocks:', o.directory.num_blocks,
                  '\n    directory block length:', len(o.directory.block.data),
                  '\n    length of blocks found in OPUSFile:', len(get_all_blocks(o)))
    # ----------------------------------------------------------------------------------------------
    # Check that `OPUSFile.directory.blocks` are clear (should be sorted elsewhere, otherwise redundant)
    redundant = [d for d in opus_data if d and len(d.directory.blocks) > 0]
    print('.' * width)
    if len(redundant) == 0:
        print_pass('OPUSFile.directory.blocks are all clear (No redundant data blocks in class)')
    else:
        print_fail('Redundant data blocks found in OPUSFile.directory.blocks (Memory ineffiency):')
        for o in redundant:
            print('   ', o.rel_path)
    # ----------------------------------------------------------------------------------------------
    # Checks for any OPUS files with unknown/uncategorized data blocks
    unknown = filter_data_with_unknown_blocks(opus_data)
    print('.' * width)
    if len(unknown) == 0:
        print_pass('No files with unknown/uncategorized blocks [Ignoring (0, 0, 0, 0, 0, 0) blocks]')
    else:
        print_fail('Files with unknown blocks: ' + str(len(unknown)) + '/' + str(len(opus_data)) + ' files')
        for o in unknown:
            print('   ', o.rel_path + ':', [b.type for b in o.unknown_blocks])
    # ----------------------------------------------------------------------------------------------
    # Checks for any `OPUSFile` that contains any unparsed block(s)
    bad_parse = filter_data_that_failed_to_parse(opus_data)
    print('.' * width)
    if len(bad_parse) == 0:
        print_pass('Every block was successfully parsed')
    else:
        print_fail('Some blocks were not parsed successfully:')
        for o in bad_parse:
            print('   ', o.rel_path)
    # ----------------------------------------------------------------------------------------------
    # Checks that all keys in PARAM_LABELS match the expected format (three capital letters or numbers)
    print('.' * width)
    failed_params = []
    for key, label in PARAM_LABELS.items():
        if not re.match('[A-Z0-9_][A-Z0-9_][A-Z0-9_]', key) or type(label) is not str:
            failed_params.append((key, label))
    if len(failed_params) > 0:
        print_fail('PARAM_LABELS contains incorrectly formatted key(s) or label(s): ' + str(failed_params))
    if len(PARAM_LABELS.keys()) < 2000:
        print_fail('Expected over 2000 PARAM_LABEL (key, label) pairs. Only found: ' + str(len(PARAM_LABELS.keys())))
    elif len(failed_params) == 0:
        print_pass('All keys and labels in PARAM_LABELS are formatted properly')
    # ==============================================================================================
    # File Statistics
    print('_' * width)
    _print_centered('Test File Statistics', width=width)
    data_keys = set()
    for o in opus_data:
        for key in o.all_data_keys:
            data_keys.add(key)
    print('Data Keys (' + str(len(data_keys)) + '):', '  '.join(sorted(list(data_keys))))


import re

from brukeropus.file.block import FileBlock

__docformat__ = "google"


class Report:
    '''Class containing data from an OPUS file report block.

    OPUS files may contain data in the form of a report block.  The report block format is generic, and can be used to
    store a variety of data types. The entire contents of a report can be viewed by printing the class to the console: 
    print(report). All of the data in the report is accessible through class attributes and indexing.

    Args:
        block: parsed `FileBlock` instance of a report block

    Attributes:
        title: (`str`) title of report
        properties: (`dict`) top-level properties of the report in the form of key, val pairs
        table: (`ReportTable`) tabular data with header and title stored in the top-level of the report
        sub: (`list`(`ReportTable`)) list of subreports (tabular data) stored in the report

    Indexing:
        The class supports indexing as a shortcut to accessing certain data within the report.
        int: subreports can be accessed by indexing an instance of a report with an integer number.
            report[0] == report.sub[0] (i.e. access first subreport)
        str: The values of report properties can be accessed by indexing an instance of a report with a string (case
            insensitive). report['version'] == report.properties['Version'] (i.e. access 'version' property of report)
    '''
    def __getitem__(self, item):
        if type(item) is int and hasattr(self, 'sub'):
            return self.sub[item]
        elif type(item) is str and item.lower() in [k.lower() for k in self.properties.keys()]:
            key = [k for k in self.properties.keys() if k.lower() == item.lower()][0]
            return self.properties[key]
        else:
            raise KeyError(str(item) + ''' not a valid key. Key can be integer (to reference subreport) or a key from
                           .properties.keys()''')

    def __init__(self, block: FileBlock):
        self.block = block
        data = self.block.data
        self.title = data['header']['tit']
        self.table = ReportTable(info=data['info'], data=data['data'], title=data['header']['e00'])
        if 'h00' in data['info'].keys():
            self.properties = {
                data['info'][key]: 
                    data['info']['v' + key[1:]] for key in data['info'].keys() if re.search('h[0-9][0-9]', key)
            }
        self.sub = []
        if 'subreports' in data.keys():
            for i, subreport in enumerate(data['subreports']):
                title = data['info']['u' + f'{i:02}']
                self.sub.append(ReportTable(title=title, **subreport))
        self._data = data
        self.block.data = None

    def __repr__(self):
        contents = []
        if hasattr(self, 'properties'):
            contents.append(f'properties ({len(self.properties)})')
        if hasattr(self, 'table'):
            contents.append('table (' + f'{self.table.num_cols}x{self.table.num_rows})')
        if hasattr(self, 'sub'):
            contents.append(f'sub ({len(self.sub)})')
        return f'Report({self.title}: ' + ', '.join(contents) + ')'

    def __str__(self, width=100, max_col_width=25, float_digits=3):
        lines = ['=' * width, f'{self.title:^{width}}', '_' * width]
        if hasattr(self, 'properties'):
            key_width = max(len(key) for key in self.properties.keys())
            for k, v in self.properties.items():
                lines.append(f'{k:>{key_width}}' + ': ' + str(v))
            lines.append('\n')
        table_str = self.table.__str__(width=width, max_width=max_col_width, float_digits=float_digits)
        lines = lines + table_str.splitlines()
        if hasattr(self, 'sub'):
            for i, s in enumerate(self.sub):
                lines.append('Subreport ' + str(i))
                s_str = s.__str__(width=width, max_width=max_col_width, float_digits=float_digits)
                lines = lines + s_str.splitlines()
        lines.append('=' * width)
        return '\n'.join(lines)


class ReportTable:
    '''Class containing table data from an OPUS file report block.

    OPUS file report blocks generally contain one or more tables of data. This class is used to store and provide an
    interface for accessing that tabular data.

    Attributes:
        title: (str) title of the table
        num_cols: (int) number of columns in the table
        num_rows: (int) number of rows in the table
        header: (list(str)) list of header labels for each row of data
        values: (list(list(vals))) 2D list of tabular data. Can be indexed: values[row][col]

    Indexing:
        The class supports indexing as a shortcut to accessing certain data within the report.
        int: indexing the table with an integer returns the corresponding row of values in the table.
            table[0] == table.values[0] (i.e. first row of data in the table)
        str: indexing the table with a string returns the row of data with matching header label (case insensitive).
            table['type'] == table.values[table.header.index('Type')]
    '''
    def __getitem__(self, item):
        if type(item) is int:
            return self.values[item]
        elif type(item) is str and item.lower() in [h.lower() for h in self.header]:
            idx = [i for i, h in enumerate(self.header) if h.lower() == item.lower()][0]
            return self.values[idx]
        else:
            raise KeyError(str(item) + ' not a valid key. For a list of valid keys, use: .headers')

    def __init__(self, info: dict, data: list, title: str):
        self._info = info
        self._data = data
        self.title = title
        self.num_cols = len(data)
        self.num_rows = len(data[0])
        self.header = [info['s' + f'{row:02}'] for row in range(self.num_rows)]
        self.values = [list(row) for row in zip(*data)]

    def __repr__(self):
        return self.title + ' (table): cols: ' + str(self.num_cols) + ', rows: ' + str(self.num_rows)

    def _str(self, val, max_width=25, float_digits=3):
        if type(val) is float:
            s = f'{val:.6g}'
        else:
            s = str(val)
        if len(s) > max_width:
            s = s[:max_width - 3] + '...'
        return s

    def __str__(self, width=100, max_width=25, float_digits=3):
        label = self.__repr__()[:width]
        lines = [label, '-' * width]
        h_width = max(len(h) for h in self.header)
        vals_width = max([width - h_width - 2, 20])
        if self.num_cols == 1:
            col_widths = [vals_width]
        else:
            col_widths = [max([len(self._str(row[col])) for row in self.values]) for col in range(self.num_cols)]
            if sum(col_widths) > vals_width:
                col_widths = [int(min([(vals_width - 2 * (self.num_cols - 1)) // self.num_cols, c]))
                              for c in col_widths]
        for i, h in enumerate(self.header):
            lines.append(f'{h:>{h_width}}' + ': ' + 
                         '  '.join([f'{self._str(v, max_width=col_widths[j], float_digits=float_digits):<{col_widths[j]}}'
                                    for j, v in enumerate(self.values[i])]))
        lines.append('-' * width)
        return '\n'.join(lines)



import numpy as np

from brukeropus.file.block import FileBlock
from brukeropus.file.params import Parameters


__docformat__ = "google"


class Data:
    '''Class containing array data and associated parameter/metadata from an OPUS file.

    Args:
        data_block: parsed `FileBlock` instance of a data block
        data_status_block: `parsed FileBlock` instance of a data status block which contains metadata about the data
            block. This block is a parameter block.
        key: attribute name (string) assigned to the data
        vel: mirror velocity setting for the measurement (from param or rf_param block as appropriate)

    Attributes:
        params: `Parameter` class with metadata associated with the data block such as first x point: `fxp`, last x
            point: `lxp`, number of points: `npt`, date: `dat`, time: `tim` etc.
        y: 1D `numpy` array containing y values of data block
        x: 1D `numpy` array containing x values of data block. Units of x array are given by `dxu` parameter.
        label: human-readable string label describing the data block (e.g. Sample Spectrum, Absorbance, etc.)
        key: attribute name (string) assigned to the data
        vel: mirror velocity setting for the measurement (used to calculate modulation frequency)
        block: data `FileBlock` used to generate the `Data` class
        blocks: [data, data_status] `FileBlock` used to generate the `Data` class

    Extended Attributes:
        **wn:** Returns the x array in wavenumber (cm⁻¹) units regardless of what units the x array was originally
            saved in. This is only valid for spectral data blocks such as sample, reference, transmission, etc., not
            interferogram or phase blocks.
        **wl:** Returns the x array in wavelength (µm) units regardless of what units the x array was originally
            saved in. This is only valid for spectral data blocks such as sample, reference, transmission, etc., not
            interferogram or phase blocks.
        **f:** Returns the x array in modulation frequency units (Hz) regardless of what units the x array was
            originally saved in. This is only valid for spectral data blocks such as sample, reference, transmission,
            etc., not interferogram or phase blocks.
        **datetime:** Returns a `datetime` class of when the data was taken (extracted from data status parameter
            block).  
        **xxx:** the various three char parameter keys from the `params` attribute can be directly called from the 
            `Data` class for convenience. Common parameters include `dxu` (x units), `mxy` (max y value), `mny` (min y
            value), etc.
    '''
    __slots__ = ('key', 'params', 'y', 'x', 'label', 'vel', 'block', 'blocks')

    def __init__(self, data_block: FileBlock, data_status_block: FileBlock, key: str, vel: float):
        self.key = key
        self.params = Parameters(data_status_block)
        if data_block.is_compact_data():
            y = data_block.data[-self.params.npt:]
        else:
            y = data_block.data
        self.y = self.params.csf * y[:self.params.npt]    # Trim extra values on some spectra
        self.x = np.linspace(self.params.fxv, self.params.lxv, self.params.npt)
        self.label = data_block.get_label()
        self.vel = vel
        data_block.data = None
        self.block = data_block
        self.blocks = self.params.blocks + [self.block]

    def __getattr__(self, name):
        if name.lower() == 'wn' and self.params.dxu in ('WN', 'MI', 'LGW'):
            return self._get_wn()
        elif name.lower() == 'wl' and self.params.dxu in ('WN', 'MI', 'LGW'):
            return self._get_wl()
        elif name.lower() == 'f' and self.params.dxu in ('WN', 'MI', 'LGW'):
            return self._get_freq()
        elif name.lower() in self.params.keys():
            return getattr(self.params, name.lower())
        elif name == 'datetime':
            return self.params.datetime
        else:
            text = str(name) + ' is not a valid attribute for Data: ' + str(self.key)
            raise AttributeError(text)

    def _get_wn(self):
        if self.params.dxu == 'WN':
            return self.x
        elif self.params.dxu == 'MI':
            return 10000. / self.x
        elif self.params.dxu == 'LGW':
            return np.exp(self.x)

    def _get_wl(self):
        if self.params.dxu == 'WN':
            return 10000. / self.x
        elif self.params.dxu == 'MI':
            return self.x
        elif self.params.dxu == 'LGW':
            return 10000 / np.exp(self.x)

    def _get_freq(self):
        vel = 1000 * np.float(self.vel) / 7900  # cm/s
        return vel * self.wn


class DataSeries(Data):
    '''Class containing a data series (3D specra) and associated parameter/metadata from an OPUS file.

    Args:
        data_block: parsed `FileBlock` instance of a data block
        data_status_block: `parsed FileBlock` instance of a data status block which contains metadata about the data
            block. This block is a parameter block.
        key: attribute name (string) assigned to the data
        vel: mirror velocity setting for measurement (from param or rf_param block as appropriate)

    Attributes:
        params: `Parameter` class with metadata associated with the data block such as first x point: `fxp`, last x
            point: `lxp`, number of points: `npt`, date: `dat`, time: `tim` etc.
        y: 2D numpy array containing y values of data block
        x: 1D numpy array containing x values of data block. Units of x array are given by `.dxu` attribute.
        num_spectra: number of spectra in the series (i.e. length of y)
        store_table: list if tuples that represent start and end indexes of saved spectra (useful for skipped spectra)
        label: human-readable string label describing the data block (e.g. Sample Spectrum, Absorbance, etc.)
        key: attribute name (string) assigned to the data
        vel: mirror velocity setting for measurement (used to calculate modulation frequency)
        block: data `FileBlock` used to generate the `DataSeries` class
        blocks: [data, data_status] `FileBlock` used to generate the `DataSeries` class

    Extended Attributes:
        **wn:** Returns the x array in wavenumber (cm⁻¹) units regardless of what units the x array was originally saved
            in. This is only valid for spectral data blocks such as sample, reference, transmission, etc., not
            interferogram or phase blocks.  
        **wl:** Returns the x array in wavelength (µm) units regardless of what units the x array was originally saved
            in. This is only valid for spectral data blocks such as sample, reference, transmission, etc., not
            interferogram or phase blocks.  
        **datetime:** Returns a `datetime` class of when the data was taken (extracted from data status parameter
            block).  
        **xxx:** the various three char parameter keys from the `params` attribute can be directly called from the data
            class for convenience. Several of these parameters return arrays, rather than singular values because they
            are recorded for every spectra in the series, e.g. `npt`, `mny`, `mxy`, `srt`, 'ert', `nsn`.
    '''
    __slots__ = ('key', 'params', 'y', 'x', 'label', 'vel', 'block', 'blocks', 'num_spectra', 'version', 'offset', 'num_blocks', 'data_size', 'info_size', 'store_table')

    def __init__(self, data_block: FileBlock, data_status_block: FileBlock, key: str, vel: float):
        self.key = key
        self.params = Parameters(data_status_block)
        data = data_block.data
        self.y = data['y'][:, :self.params.npt]    # Trim extra values on some spectra
        self.x = np.linspace(self.params.fxv, self.params.lxv, self.params.npt)
        self.num_spectra = len(self.y)
        for key in ['version', 'offset', 'num_blocks', 'data_size', 'info_size', 'store_table']:
            setattr(self, key, data[key])
        for key, val in data.items():
            if key not in ['version', 'offset', 'num_blocks', 'data_size', 'info_size', 'store_table']:
                self.params._params[key] = val
        self.label = data_block.get_label()
        self.vel = vel
        data_block.data = None
        self.block = data_block
        self.blocks = self.params.blocks + [self.block]



import os, struct, errno
import numpy as np
from brukeropus.file.constants import STRUCT_3D_INFO_BLOCK, SUBREPORT_TYPE_FMT


__docformat__ = "google"


def read_opus_file_bytes(filepath) -> bytes:
    '''Returns `bytes` of an OPUS file specified by `filepath` (or `None`).

    Function determines if `filepath` points to an OPUS file by reading the first four bytes which are always the same
    for OPUS files.  If `filepath` is not a file, or points to a non-OPUS file, the function returns `None`.  Otherwise
    the function returns the entire file as raw `bytes`.

    Args:
        filepath (str or Path): full filepath to OPUS file

    Returns:
        **filebytes (bytes):** raw bytes of OPUS file or `None` (if filepath does not point to an OPUS file)
    '''
    filebytes = None
    if os.path.isfile(filepath):
        with open(filepath, 'rb') as f:
            try:
                first_four = f.read(4)
                if first_four == b'\n\n\xfe\xfe':
                    filebytes = first_four + f.read()
            except:
                pass # Empty file (or file with fewer than 4 bytes)
    else:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filepath)
    return filebytes


def get_block_type(type_int: int) -> tuple:
    '''Converts an int32 block type code to a six-integer tuple `block_type`.

    This function is used to decode the `type_int` from the directory block of an OPUS file into a tuple of integers.
    Each integer in the tuple provides information about the associated data block.

    Args:
        type_int: 32-bit integer decoded from file directory block

    Returns:
        **block_type (tuple):** six-integer tuple which specifies the block type
    '''
    type_bit_str = format(type_int, '#034b')  # binary representation as string
    block_type = (
        int(type_bit_str[-2:], 2),
        int(type_bit_str[-4:-2], 2),
        int(type_bit_str[-10:-4], 2),
        int(type_bit_str[-17:-10], 2),
        int(type_bit_str[-19:-17], 2),
        int(type_bit_str[-22:-19], 2)
    )
    return block_type


def decode_str(size: int, blockbytes: bytes, offset: int) -> str:
    '''Decode string that is packed as bytes in `blockbytes` starting from `offset`.

    Strings are frequently stored in OPUS files with a size designation that is larger than the actual string. The end
    of the string is designated by a terminator byte: b'\x00'. This function unpacks the string using the size
    designator, truncates at the terminator byte if found, and decodes as "latin-1"

    Args:
        size: size (number of bytes) of the string
        blockbytes: raw bytes of an OPUS file block
        offset: offset location where string begins in blockbytes

    Returns:
        string: decoded string
    '''
    fmt = '<' + str(size) + 's'
    try:
        val = struct.unpack_from(fmt, blockbytes, offset)[0]
        x00_pos = val.find(b'\x00')
        if x00_pos != -1:
            val = val[:x00_pos].decode('latin-1')
        else:
            val = val.decode('latin-1')
    except Exception as e:
        val = 'Failed to decode: ' + str(e)
    return val


def parse_header(filebytes: bytes) -> tuple:
    '''Parses the OPUS file header.

    The header of an OPUS file contains some basic information about the file including the version number, location of
    the directory block, and number of blocks in the file. This header is first to be parsed as it specifies how to
    read the file directory block (which contains information about each block in the file)

    Args:
        filebytes: raw bytes of OPUS file (all bytes)

    Returns:
        **header_info (tuple):**  
            (  
                **version (float64):** program version number as a floating-point date (later versions always greater)  
                **directory_start (int32):** pointer to start location of directory block (number of bytes)  
                **max_blocks (int32):** maximum number of blocks supported by the directory block (this should only be
                    relevant when trying to edit an OPUS file, i.e. when adding data blocks to a file)  
                **num_blocks (int32):** total number of blocks in the opus file  
            )
    '''
    version = struct.unpack_from('<d', filebytes, 4)[0]
    directory_start = struct.unpack_from('<i', filebytes, 12)[0]
    max_blocks = struct.unpack_from('<i', filebytes, 16)[0]
    num_blocks = struct.unpack_from('<i', filebytes, 20)[0]
    return version, directory_start, max_blocks, num_blocks


def parse_directory(blockbytes: bytes) -> list:
    '''Parses directory block of OPUS file and returns a list of block info tuples: (type, size, start).

    The directory block of an OPUS file contains information about every block in the file. The block information is
    stored as three int32 values: `type_int`, `size_int`, `start`.  `type_int` is an integer representation of the block
    type. The bits of this `type_int` have meaning and are parsed into a tuple using `get_block_type`. The `size_int` is
    the size of the block in 32-bit words. `start` is the starting location of the block (in number of bytes).

    Args:
        blockbytes: raw bytes of an OPUS file directory block

    Returns:
        **blocks (list):** list of block_info tuples
            **block_info (tuple):**
                (  
                    **block_type (tuple):** six-integer tuple which specifies the block type (see: `get_block_type`)  
                    **size (int):** size (number of bytes) of the block  
                    **start (int):** pointer to start location of the block (number of bytes)
                )
    '''
    loc = 0
    blocks = []
    while loc < len(blockbytes):
        type_int, size_int, start = struct.unpack_from('<3i', blockbytes, loc)
        loc = loc + 12
        if start > 0:
            block_type = get_block_type(type_int)
            size = size_int*4
            blocks.append((block_type, size, start))
        else:
            break
    return blocks


def parse_params(blockbytes: bytes) -> dict:
    '''Parses the bytes in a parameter block and returns a dict containing the decoded keys and vals.

    Parameter blocks are in the form: `XXX`, `dtype_code`, `size`, `val`.  `XXX` is a three char abbreviation of the
    parameter (key). The value of the parameter is decoded according to the `dtype_code` and size integers to be either:
    `int`, `float`, or `string`.

    Args:
        blockbytes: raw bytes of an OPUS file parameter block

    Returns:
        **items (tuple):** (key, value) pairs where key is three char string (lowercase) and value can be `int`, `float`
            or `string`.
    '''
    loc = 0
    params = dict()
    while loc < len(blockbytes):
        key = blockbytes[loc:loc + 3].decode('utf-8')
        if key == 'END':
            break
        dtype_code, val_size = struct.unpack_from('<2h', blockbytes[loc + 4:loc + 8])
        val_size = val_size * 2
        if dtype_code == 0:
            val = struct.unpack_from('<i', blockbytes, loc + 8)[0]
        elif dtype_code == 1:
            val = struct.unpack_from('<d', blockbytes, loc + 8)[0]
        else:
            val = decode_str(val_size, blockbytes, loc + 8)
        params[key.lower()] = val
        loc = loc + val_size + 8
    return params


def get_dpf_dtype_count(dpf: int, size: int) -> tuple:
    '''Returns numpy dtype and array count from the data point format (dpf) and block size (in bytes).

    Args:
        dpf: data point format integer stored in data status block.
            dpf = 1 -> array of float32
            dpf = 2 -> array of int32
        size: Block size in bytes.

    Returns:
        **dtype (numpy.dtype):** `numpy` dtype for defining an `ndarray` to store the data
        **count (int):** length of array calculated from the block size and byte size of the dtype.
    '''
    if dpf == 1:
        dtype = np.float32
        count = round(size/4)
    elif dpf == 2:
        dtype = np.int32
        count = round(size/4)
    else:
        print('Unknown Data Point Format Requested:', dpf, '[using default: `float32`]')
        dtype = np.float32
        count = round(size/4)
    return dtype, count


def parse_data(blockbytes: bytes, dpf: int = 1) -> np.ndarray:
    '''Parses the bytes in a data block and returns a `numpy` array.

    Data blocks contain no metadata, only the y-values of a data array. Data arrays include: single-channel sample,
    reference, phase, interferograms, and a variety of resultant data (transmission, absorption, etc.).  Every data
    block should have a corresponding data status parameter block which can be used to generate the x-array values for
    the data block. The data status block also specifies the data type of the data array with the `DPF` parameter. It
    appears that OPUS currently exclusively stores data blocks as 32-bit floats, but has a reservation for 32-bit
    integers when `DPF` = 2.

    Args:
        blockbytes: raw bytes of data block
        dpf: data-point-format integer stored in corresponding data status block.

    Returns:
        **y_array (numpy.ndarray):** `numpy` array of y values contained in the data block
    '''
    dtype, count = get_dpf_dtype_count(dpf=dpf, size=len(blockbytes))
    return np.frombuffer(blockbytes, dtype=dtype, count=count)


def parse_data_compact(blockbytes: bytes, npt: int, dpf: int = 1) -> np.ndarray:
    '''Parses the bytes in a data compact block and returns a `numpy` array.

    Some data blocks are stored in the "Compact" block format that includes some metadata that preceeds the raw data. At
    this time, the metadata is ignored, and the compact spectra is extracted from the last bytes of the block that fit
    the known array size (specified with "npt" in corresponding data status block).

    Args:
        blockbytes: raw bytes of data block
        npt: number of data points in the spectra (from data status block)
        dpf: data-point-format integer stored in corresponding data status block.

    Returns:
        **y_array (numpy.ndarray):** `numpy` array of y values contained in the data block
    '''
    dtype, count = get_dpf_dtype_count(dpf=dpf, size=len(blockbytes))
    return np.frombuffer(blockbytes, dtype=dtype, count=count)


def parse_data_series(blockbytes: bytes, dpf: int = 1) -> dict:
    '''Parses the bytes in a 3D data block (series of spectra) and returns a data `dict` containing data and metadata.

    3D data blocks are structured differently than standard data blocks. In addition to the series of spectra, they
    include metadata for each of the spectrum.  This function returns a `dict` containing all the extracted information
    from the data block.  The series spectra is formed into a 2D array while metadata captured for each spectra is
    formed into a 1D array (length = number of spectral measurements in the series).

    Args:
        blockbytes: raw bytes of the data series block
        dpf: data-point-format integer stored in corresponding data status block.

    Returns:
        **data_dict (dict):** `dict` containing all extracted information from the data block  
            {  
                **version:** file format version number (should be 0)  
                **num_blocks:** number of sub blocks; each sub block features a data spectra and associated metadata  
                **offset:** offset in bytes to the first sub data block  
                **data_size:** size in bytes of each sub data block  
                **info_size:** size in bytes of the metadata info block immediately following the sub data block  
                **store_table:** run numbers of the first and last blocks to keep track of skipped spectra  
                **y:** 2D `numpy` array containing all spectra (C-order)  
                **metadata arrays:** series of metadata arrays in 1D array format (e.g. `npt`, `mny`, `mxy`, `ert`).
                    The most useful one is generally `ert`, which can be used as the time axis for 3D data plots.
            }
    '''
    header = struct.unpack_from('<6i', blockbytes, 0)
    data = {
        'version': header[0],
        'num_blocks': header[1],
        'offset': header[2],
        'data_size': header[3],
        'info_size': header[4],
    }
    data['store_table'] = [struct.unpack_from('<2i', blockbytes, 24 + i * 8) for i in range(header[5])]
    dtype, count = get_dpf_dtype_count(dpf, data['data_size'])
    data['y'] = np.zeros((data['num_blocks'], count), dtype=dtype)
    for entry in STRUCT_3D_INFO_BLOCK:
        data[entry['key']] = np.zeros((data['num_blocks']), dtype=entry['dtype'])
    offset = data['offset']
    for i in range(data['num_blocks']):
        data['y'][i] = np.frombuffer(blockbytes[offset:], dtype=dtype, count=count)
        offset = offset + data['data_size']
        info_vals = struct.unpack_from('<' + ''.join([e['fmt'] for e in STRUCT_3D_INFO_BLOCK]), blockbytes, offset)
        for j, entry in enumerate(STRUCT_3D_INFO_BLOCK):
            data[entry['key']][i] = info_vals[j]
        offset = offset + data['info_size']
        if offset >= len(blockbytes):
            num_spectra = i
            break # Not all blocks are necessarily stored (see Store Table)
        num_spectra = i
    data['y'] = data['y'][:num_spectra]
    for entry in STRUCT_3D_INFO_BLOCK:
        data[entry['key']] = data[entry['key']][:num_spectra]
    return data


def parse_text(block_bytes: bytes, encoding='utf-8', backup_encoding='utf-16') -> str:
    '''Parses and OPUS file block as text (e.g. history or file-log block).

    The history (aka file-log) block of an OPUS file contains some information about how the file was generated and
    edits that have been performed on the file.  This function parses the text block but does not take any steps to
    parameterizing what is contained in the text.  The history block is generally not needed to retrieve the file data
    and metadata, but might be useful for inspecting the file.

    Args:
        blockbytes: raw bytes of the text block (e.g. history or file-log)
        encoding: string represting text encoding type. Can be set to "utf-16" for chinese character support
        backup_encoding: if default encoding fails, this encoding string will be attempted.

    Returns:
        text: string of text contained in the file block.
    '''
    byte_string = struct.unpack('<' + str(len(block_bytes)) + 's', block_bytes)[0]
    byte_strings = byte_string.split(b'\x00')
    strings = []
    for entry in byte_strings:
        if entry != b'':
            try:
                strings.append(entry.decode(encoding))
            except Exception:
                try:
                    strings.append(entry.decode(backup_encoding))
                except Exception:
                    try:
                        strings.append(entry.decode('latin-1'))
                    except Exception as e:
                        strings.append('<Decode Exception>: ' + str(e))
    return '\n'.join(strings)


def parse_subreport(subreport_bytes: bytes) -> dict:
    '''Parses the bytes of a subreport and returns the extracted data as a dictionary

    Subreports are contained within a report block (e.g. Multi-Evaluation Test Report). A report can contain multiple
    subreports, and they generally follow a table format. This sub-block is organized with a mini parameter block
    followed by packed data. The mini parameter block contains information about how to read the packed data:
        nco: number of columns
        nln: number of rows
        siz: size of mini parameter block (number of bytes)
        src: size in bytes of entire row of data (offset for extracting column data from row 2, 3 ...)
        f00, f01 ... fxx: start position of data in column 0, 1 ... xx (relative to end of mini param block)
        t00, t01 ... txx: integer representing type of data (e.g. int32, float32, float64, str, etc.)
        s00, s01 ... sxx: column header label
        p00, p01 ... pxx: formatting string for numbers in column 0, 1 ... xx (not included for every column)

    Args:
        subreport_bytes: raw bytes of the subreport. Needs to start precisely where subreport begins, but can include
        data beyond the end of the subreport (i.e. end of subreport does not need to be determined a priori).

    Returns:
        **subreport (dict):** `dict` containing subreport data and extraction/formatting parameters  
            {  
                **info:** `dict` of parameters extracted directly from subreport that describes how to read the data
                    table and provides some basic metadata about the table (e.g. column header labels).
                **data:** `list` of lists of data (table format) contained in the subreport
            }
    '''
    info = parse_params(subreport_bytes)
    data = []
    for row in range(info['nln']):
        data.append([])
        for col in range(info['nco']):
            offset = info['siz'] + row * info['src'] + info['f' + f'{col:02}']
            type_int = info['t' + f'{col:02}']
            if col < info['nco'] - 1:
                size = min([type_int - 1000, info['f' + f'{col + 1:02}'] - info['f' + f'{col:02}']])
            else:
                size = info['src'] - info['f' + f'{col:02}']
            if type_int > 1000:
                val = decode_str(size, subreport_bytes, offset)
            elif type_int in SUBREPORT_TYPE_FMT.keys():
                fmt = SUBREPORT_TYPE_FMT[type_int]
                val = struct.unpack_from(fmt, subreport_bytes, offset)[0]
            else:
                val = subreport_bytes[offset:offset + size]
            data[row].append(val)
    return {'info': info, 'data': data}


def parse_report(blockbytes: bytes) -> dict:
    '''Parses the report block of an OPUS file, such as Multi-Evaluation test reports, returning the report as a dict.

    Report blocks are formatted in a very general way, potentially enabling a variety of different report structures.
    This algorithm was developed using several OPUS files with a variety of different Multi-Evaluation Test Reports.
    It is possible that other classes of test reports could be generated by OPUS that might change the structure, but
    the overal organization and decoding methods should be similar.  In particular, the report block structure might
    support multiple reports, but no such file has been available for testing to date.  This algorithm will extract a
    single report and all the associated subreports.

    Report blocks start with a mini parameter block that begins after the 12th byte.  It contains the following:
        tit: Title of the report
        f00: Starting position of the report summary data
        Known unused parameters: bid, nrp, siz, e00, z00
    This is followed by the report summary. For a multi-evaluation test report, this is a pair of tables summarizing the
    methods applied to the spectrum.  It also specifies the number of subreports that follow, and the starting position
    and title of each subreport. Some of the keys in this parameter set are described in the `parse_subreport` method.
    Other parameters in the report summary include:
        sub: Number of subreports
        h00, h01 ... hxx: header labels of first summary table
        v00, v01 ... vxx: corresponding values of first summary table
        g00, g01 ... gxx: starting positions of each subreport relative to the start of this report summary
        u00, u02 ... uxx: titles of each subreport
    It should be noted that the only class of reports used for testing this algorithm were a variety of multi-evaluation
    test reports. It is possible there are other similar report blocks OPUS writes that follow a similar structure but
    could vary in some way that is incompatible with some of the assumptions made by this algorithm.

    Args:
        blockbytes: raw bytes of an OPUS file report block

    Returns:
        **report (dict):** `dict` containing report and subreport data 
            {
                **header:** `dict` of parameters from first mini param block
                **info:** `dict` of parameters extracted from second mini param block
                **data:** `list` of data that comprises second summary table
                **subreports:** `list` of subreport `dict`s beneath the main report.
            }
    '''
    header_ints = struct.unpack('<3i', blockbytes[:12])
    header_info = parse_params(blockbytes[12:])
    header_info['ints'] = header_ints
    report_info = parse_subreport(blockbytes[header_info['f00']:])
    report = {'header': header_info, 'info': report_info['info'], 'data': report_info['data']}
    subreports = []
    for idx in range(report_info['info']['sub']):
        offset = header_info['f00'] + report_info['info']['g' + f'{idx:02}']
        subreports.append(parse_subreport(blockbytes[offset:]))
    report['subreports'] = subreports
    return report



from brukeropus.file.constants import TYPE_CODE_LABELS, CODE_3_ABR
from brukeropus.file.parse import (parse_directory,
                                   parse_params,
                                   parse_data,
                                   parse_data_series,
                                   parse_text,
                                   parse_report)


__docformat__ = "google"


class BlockType(tuple):
    '''Six-integer tuple representing the category (type) of block within an OPUS file.

    Each block in an OPUS file is categorized with six integers, for example (3, 1, 1, 2, 0, 0). This class stores the
    integers as a `tuple`, but extends the `tuple` class to provide a few useful functions/attributes.

    Args:
        block_type: six integers found in the OPUS file directory that describe the block type.

    Attributes:
        label: human-readable label that describes the block category
    '''

    def get_label(self):
        '''Converts a six-integer tuple block type into a human readable label.

        This package includes the majority of type codes that OPUS uses, but in the event a type code label is not known,
        this function will return: "Unknown 0 4" where the first number is the position index, and the second is the
        unknown value integer.

        Args:
            block_type: six integer tuple found in the OPUS file directory that describes the block type

        Returns:
            label (str): human-readable string label
        '''
        labels = [self._get_sub_type_label(idx) for idx in range(len(self)) if self[idx] > 0
                  and self._get_sub_type_label(idx) != '']
        if len(labels) > 0:
            return ' '.join(labels)
        else:
            return 'Undefined'

    def _get_sub_type_label(self, pos_idx: int):
        '''Returns the sub-type label of a file block given the position index and value of the type code.

        Args:
            pos_idx: positional index of the type code (0 - 5)

        Returns:
            label (str): human-readable string label that describes the type code at that index.
        '''
        if pos_idx == 3:
            channels, type_idx = divmod(self[3], 32)
        else:
            channels, type_idx = 0, self[pos_idx]
        try:
            label = TYPE_CODE_LABELS[pos_idx][type_idx]
        except KeyError:
            label = 'Unknown ' + str(pos_idx) + ' ' + str(type_idx)
        if channels > 0:
            label = label + ' ' + str(channels + 1) + '-Channel'
        return label

    def get_aligned_tuple_str(self, pad=1):
        return f'{self[0]}' + f'{self[1]:2}' + f'{self[2]:3}' + f'{self[3]:3}' + f'{self[4]:2}' + f'{self[5]:2}'

    def __repr__(self):
        return 'BlockType((' + ', '.join([str(i) for i in self]) + '))'

    def __str__(self):
        return self.get_aligned_tuple_str() + '   ' + self.get_label()

    def __new__(cls, iterable):
        instance = super().__new__(cls, iterable)
        if len(instance) != 6 or any(type(i) != int for i in instance):
            raise ValueError('BlockType input must be a 6-integer iterable, but a value of:' + str(iterable) + ' was given')
        return instance

    def __init__(self, iterable):
        super().__init__()
        self.label = self.get_label()


class FileBlock:
    '''Generic OPUS file block.

    This class initializes from the block info stored in the file directory (type, size, and start location) as well as
    the raw bytes from the file (which can subsequently be parsed).

    Args:
        filebytes: raw bytes of the file
        block_type: six integer tuple that describes the type of data in the file block
        size: size of block in number of bytes
        start: pointer to start location of the block within the file
        debug: whether to read the block in debug mode (default: False)

    Attributes:
        type: six integer tuple that describes the type of data in the file block
        size: size of block in number of bytes
        start: pointer to start location of the block within the file
        bytes: raw bytes of file block (set to zero bytes if successfully parsed unless in debug mode)
        data: parsed data if successful. Could be: `list`, `str`, `np.ndarray` or `dict` depending on the block type.
        parser: name of parsing function if parsing was successful
    '''

    __slots__ = ('type', 'size', 'start', 'bytes', 'data', 'parser', 'parse_error', 'keys', 'debug')

    def __init__(self, filebytes: bytes, block_type: tuple, size: int, start: int, debug=False):
        self.bytes = filebytes[start: start + size]
        self.type = BlockType(block_type)
        self.size = size
        self.start = start
        self.debug = debug
        self.data = None
        self.parser = None
        self.parse_error = None

    def __str__(self):
        label = self.get_label()
        return 'FileBlock: ' + label + ' (size: ' + str(self.size) + ' bytes; start: ' + str(self.start) + ')'

    def _try_parser(self, parser):
        try:
            self.data = parser(self.bytes)
            if type(self.data) is dict:
                self.keys = list(self.data.keys())
            self._clear_parsed_bytes(parser=parser)
        except Exception as e:
            self.parse_error = 'Error parsing (' + parser.__name__ + '): ' + str(e)

    def _clear_parsed_bytes(self, parser):
        '''Clear raw bytes that have been parsed (and log the parser for reference)'''
        self.parser = parser.__name__
        if not self.debug:
            self.bytes = b''

    def is_data_status(self):
        '''Returns True if `FileBlock` is a data status parameter block'''
        return self.type[2] == 1

    def is_rf_param(self):
        '''Returns True if `FileBlock` is a parameter block associated with the reference measurement (not including
        data status blocks)'''
        return self.type[2] > 1 and self.type[1] == 2

    def is_param(self):
        '''Returns True if `FileBlock` is any parameter block (could be data status, rf param, sample param, etc.)'''
        return self.type[2] > 0 or self.type == (0, 0, 0, 0, 0, 1)

    def is_sm_param(self):
        '''Returns True if `FileBlock` is a parameter block associated with sample/result measurement (not including
        data status blocks)'''
        return self.is_param() and not self.is_data_status() and not self.is_rf_param()

    def is_directory(self):
        '''Returns True if `FileBlock` is the directory block'''
        return self.type == (0, 0, 0, 13, 0, 0)

    def is_file_log(self):
        '''Returns True if `FileBlock` is the file log (aka 'history') block'''
        return self.type == (0, 0, 0, 0, 0, 5)

    def is_report(self):
        '''Returns True if `FileBlock` is a test report'''
        return self.type in [(0, 0, 0, 0, 0, 2), (0, 0, 0, 0, 0, 3), (0, 0, 0, 0, 0, 4)] or (self.type[2] == 0 and self.type[3] not in [0, 13] and self.type[5] == 5)

    def is_data(self):
        '''Returns True if `FileBlock` is a 1D data block (not a data series)'''
        return self.type[2] == 0 and self.type[3] not in [0, 13] and self.type[5] not in [2, 5]

    def is_data_series(self):
        '''Returns True if `FileBlock` is a data series block (i.e. 3D data)'''
        return self.type[2] == 0 and self.type[3] not in [0, 13] and self.type[5] == 2
    
    def is_compact_data(self):
        '''Returns True if `FileBlock` is a compact 1D data block. These data blocks have metadata preceeding the data
        array (currently ignored).'''
        return self.is_data() and self.type[5] == 4

    def get_label(self):
        '''Returns a friendly string label that describes the block type'''
        return self.type.label

    def get_data_key(self):
        '''If block is a data block, this function will return a shorthand key to reference that data.

        e.g. t: transmission, a: absorption, sm: sample, rf: reference, phsm: sample phase etc. If the block is not
        a data block, it will return `None`.'''
        if self.is_data() or self.is_data_series():
            channels = self.type[3] // 32 + 1
            type_idx = self.type[3] % 32
            if type_idx in CODE_3_ABR.keys():
                key = CODE_3_ABR[type_idx]
            else:
                key = '_' + str(self.type[3])
            if self.type[1] == 1:
                key = key + 'sm'
            elif self.type[1] == 2:
                key = key + 'rf'
            elif self.type[1] > 3:
                key = key + '_' + str(self.type[1])
            if channels > 1:
                key = key + '_' + str(channels) + 'ch'
            if self.type[5] == 4:
                key = key + '_c'
            return key
        else:
            return None

    def get_parser(self):
        '''Returns the appopriate file block parser based on the type code (None if not recognized)'''
        if self.is_directory():
            return parse_directory
        elif self.is_file_log():
            return parse_text
        elif self.is_param():
            return parse_params
        elif self.is_report():
            return parse_report
        elif self.is_data_series():
            return parse_data_series
        elif self.is_data():
            return parse_data
        else:
            return None

    def parse(self):
        '''Determines the appropriate parser for the block and parses the raw bytes.  Parsed data is stored in `data`
        attribute and `bytes` attribute is set empty to save memory'''
        parser = self.get_parser()
        if parser is not None:
            self._try_parser(parser)


def is_data_status_type_match(data_block: FileBlock, data_status_block: FileBlock) -> bool:
    '''Checks if data and data status blocks are a match based soley on the block type.

    This check correctly and accurately matches blocks most of the time, but is occasionally not sufficient on its own
    (e.g. when multiple spectra of the same exact type are stored in a single file)'''
    t1 = data_status_block.type
    t2 = data_block.type
    return t1[:2] == t2[:2] and t1[3:] == t2[3:]


def is_data_status_val_match(data_block: FileBlock, data_status_block: FileBlock) -> bool:
    '''Checks if min(data) and max(data) match up with the data status parameters: MNY and MXY.

    When multiple spectra of the same type exist in a file, this is used to distinguish if the data and data status
    blocks are a good match.  This can reduce the number of duplicate matches, but is not generally sufficient to
    fully eliminate duplicate matches.

    See test file: `Test Vit C_Glass.0000_comp.0`'''
    if data_block.is_data():
        try:
            ds = data_status_block.data
            if len(data_block.data) < ds['npt']:
                return False
            else:
                y = ds['csf'] * data_block.data[:ds['npt']]
                return y.min() == ds['mny'] and y.max() == ds['mxy']
        except:
            return True # If error, can't rule out the match
    else:
        return True # Don't rule out data series at this time (no example files to test)


def is_valid_match(data_block: FileBlock, data_status_block: FileBlock) -> bool:
    '''Checks that number of points in data status are less than or equal to length of parsed data block.

    This does not apply to data series. While rare, it is occasionally necessary to remove these bad matches.

    See test file: `unreadable.0000`'''
    if data_block.is_data() and len(data_block.data) < data_status_block.data['npt']:
        return False
    else:
        return True


def pair_data_and_status_blocks(blocks: list) -> list:
    '''Takes a list of `FileBlock` and returns a list of matching (data, data status) blocks for further processing.

    All valid data blocks have an associated data status parameter block that contains y-scaling and x-axis info.
    Generally, these blocks can be easily paired with one another by using the block type. However, some files can
    contain multiple data blocks with the same exact type, which leads to duplicate matches and requires further
    inspection to accurately pair.  This function uses the following logical sequence to accurately pair these blocks:

        1. Pair by type and isolate singular matches from duplicate matches
        2. For duplicate matches, check min(data) and max(data) match data status MNY and MXY (including CSF scaling)
        3. Again isolate singular matches from duplicate matches
        4. For remaining duplicate matches, remove any matches that are already in the singular match list.
        5. Again isolate singular matches (ideally no remaining duplicate matches at this point)
        6. Remove invalid matches from the singular match list if len(data) is < data status NPT (invalid condition)
        7. Sort the matches in reverse order of where the data blocks are stored in the file (presume last is most
           recently added, and therefore final spectra which was true for limited test files).

    For the very limited test files available, this is sufficient to accurately pair all the blocks, but it seems
    plausible that it will not be sufficient for all files that have duplicate data type entries. More test files
    required for thorough testing.
    '''
    data_status = [b for b in blocks if b.is_data_status() and type(b.data) is not str]
    data = [b for b in blocks if b.is_data() or b.is_data_series() and type(b.data) is not str]
    type_matches = []
    for d in data:
        type_matches.append((d, [b for b in data_status if is_data_status_type_match(d, b)]))
    single_matches = [(m[0], m[1][0]) for m in type_matches if len(m[1]) == 1]
    multi_matches = [match for match in type_matches if len(match[1]) > 1]
    val_matches = []
    for d, matches in multi_matches:
        val_matches.append((d, [b for b in matches if is_data_status_val_match(d, b)]))
    single_matches = single_matches + [(m[0], m[1][0]) for m in val_matches if len(m[1]) == 1]
    multi_matches = [match for match in val_matches if len(match[1]) > 1]
    reduced_matches = []
    single_starts = [m[1].start for m in single_matches]
    for d, matches in multi_matches:
        reduced_matches.append((d, [b for b in matches if b.start not in single_starts]))
    single_matches = single_matches + [(m[0], m[1][0]) for m in reduced_matches if len(m[1]) == 1]
    multi_matches = [match for match in reduced_matches if len(match[1]) > 1]

    single_matches = [match for match in single_matches if is_valid_match(match[0], match[1])] # remove invalid

    single_matches.sort(key=lambda pairs: pairs[0].start, reverse=True) # last spec seems to be OPUS preference
    return single_matches



'''
The `brukeropus.control` submodule of `brukeropus` includes the `Opus` class for communicating with OPUS software. The
`Opus` class currently supports communication through the Dynamic Data Exchange (DDE) protocol.  This class can be used
to script measurement sweeps and perform various low-level operations (e.g. move mirrors, rotate polarizers, etc.). In
order to communicate with OPUS, the software must be open, logged in, and running on the same PC as `brukeropus`.
### Initializing/verifying connection to OPUS Software
```python
from brukeropus import Opus

opus = Opus()  # initialization of class automatically connects to open OPUS software
print(opus.get_version())  # prints the current OPUS software version
```
### Get information about a parameter (e.g. DTC, APT, VEL).
```python
opus = Opus()
param = 'vel'
print(opus.get_param_label(param))
print(opus.get_param_options(param))
```
### Perform a measurement sweep
```python
from brukeropus import opus, read_opus
from matplotlib import pyplot as plt

opus = Opus()  # Connects to actively running OPUS software

apt_options = opus.get_param_options('apt') # Get all valid aperture settings

for apt in apt_options[2:-2]: # Loop over all but the two smallest and two largest aperature settings
    filepath = opus.measure_sample(apt=apt, nss=10, unload=True) # Perform measurement and unload file from OPUS
    data = read_opus(filepath) # Read OPUS file from measurement
    plt.plot(data.sm.x, data.sm.y, label=apt) # Plot single-channel sample spectra
plt.legend()
plt.show()
```
For complete `Opus` documentation, see: `brukeropus.control.opus`
'''
from brukeropus.control.dde import DDEClient
from brukeropus.control.opus import *
